/*
 * Author: Marc Bevand <m.bevand@gmail.com>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "descrypt.h"
#include ".current_salt"

/*
 * Extract the value stored in bit <bitn> of each value of the array
 * <a>.
 */
unsigned long long extract_bit(const qword *a, unsigned alen, unsigned bitn)
{
    unsigned long long res = 0;
    unsigned i;

    for (i = 0; i < alen; i++)
      {
	vector unsigned int v = (vector unsigned int)a[i];
	res |= (unsigned long long)
	    (!!(spu_extract(v, bitn / 32) & (1 << (bitn % 32)))) << i;
      }
    return res;
}

static const char itoa64[] =           /* 0..63 => ascii-64 */
"./0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";

/*
 * Converts a ciphertext <ciph> and a salt <salt> to a textual
 * representation.
 */
void salt_ciph_2_str(char *dest, unsigned dlen, unsigned salt, unsigned long long ciph)
{
    int i;

    if (dlen--)
	*dest++ = itoa64[salt & 0x3f];
    if (dlen--)
	*dest++ = itoa64[(salt >> 6) & 0x3f];
    for (i = 64 - 6; dlen && i > -6; i -= 6)
      {
	if (i >= 0)
	    *dest++ = itoa64[(ciph >> i) & 0x3f];
	else
	    *dest++ = itoa64[(ciph << abs(i)) & 0x3f];
	dlen--;
      }
    if (dlen--)
	*dest++ = 0;
}

/*
 * Converts the hash {@code str} to a 12-bit salt and 64-bit ciphertext.
 */
void str_2_salt_ciph(const char *str, unsigned *salt, unsigned long long *ciph)
{
    if (strlen(str) != 13)
	fprintf(stderr, "hash of length %zd (expected 13)\n", strlen(str)),
	    exit(1);
    unsigned i = 0;
    *salt = 0;
    *ciph = 0;
    while (i < strlen(str))
      {
	const char *p = strchr(itoa64, str[i]);
	if (!p)
	    fprintf(stderr, "illegal char in hash '%s'\n", str), exit(1);
	unsigned long long valchar = p - itoa64;
	if (i < 2)
	    *salt |= valchar << (i * 6);
	else if (i < 12)
	    *ciph |= valchar << (((10 - (i - 2)) * 6) - 2);
	else
	  {
	    if (valchar & 0x3)
		printf("warning: 2 less significant bits of last base64 char "
			"are not 0: '%s'\n", str);
	    *ciph |= valchar >> 2;
	  }
	i++;
      }
}

long long pw_val(const char *p)
{
    long long ret = 0;
    unsigned i;
    for (i = 0; i < strlen(p); i++)
      {
	ret *= ('~' - ' ' + 1);
	ret += (p[i] - ' ');
      }
    return ret;
}

/*
 * Computes {@code a - b} in the printable ASCII password space.
 */
long long pw_minus(const char *a, const char *b)
{
    if (strlen(a) != strlen(b))
	fprintf(stderr, "*bug*: different length for '%s' and '%s'\n", a, b),
	    exit(1);
    return pw_val(a) - pw_val(b);
}

/*
 * Increments the prefix of length {@code prelen} of {@code pw}.
 */
void pw_inc(char *pw, unsigned prelen)
{
    if (!prelen)
	// special case, pw was empty, next pw is a single SPACE
	strcpy(pw, " ");
    else
      {
	int i = prelen - 1;
	while (i >= 0)
	  {
	    pw[i]++;
	    if (pw[i] <= '~')
		return;
	    pw[i] = ' ';
	    i--;
	  }
	if (strlen(pw) > 8)
	    fprintf(stderr, "*bug*: pw should be of length 8 max: '%s'\n", pw),
		exit(1);
	// increase the length of the pw by 1, copy up to 9 SPACEs as this
	// is a way to indicate the end of the pw space has been reached
	i = strlen(pw) + 1;
	strcpy(pw, "         ");
	pw[i] = 0;
      }
}

/*
 * Initializes {@code k} with passwords of length {@code len}.
 * Out of the passwords starting from {@code pw_cur} up to and including
 * {@code pw_last}, identify up to 128 prefixes (unique in the first
 * {@code prelen} bytes) and store them in up to 128 keys in {@code k}, fill
 * in the suffixes with the SPACE character, and increment {@code pw_cur}.
 *
 * Return 0 if {@code pw_cur} is already past {@code pw_last}, else
 * return 1.
 */
int k_load_pw_prefix_space(qword *k, char *pw_cur, const char *pw_last,
	unsigned len, unsigned prelen)
{
    // "reset" the less significant chars of the passwd (the suffix),
    // because it simplifies the algorithm in break_pw_len()
    unsigned i = prelen;
    while (i < len)
	pw_cur[i++] = ' ';
    // if pw_cur is past pw_last, return 0
    long long nr_pw;
    if (strlen(pw_cur) > strlen(pw_last) ||
	    (nr_pw = pw_minus(pw_last, pw_cur) + 1) <= 0)
	return 0;
    // count how many keys need to be initialized
    i = prelen;
    while (i++ < len)
	nr_pw = (nr_pw - 1) / ('~' - ' ' + 1) + 1;
    int nr_keys = nr_pw < 128 ? nr_pw : 128;
    // init the prefixes
    memset(k, 0, 56 * sizeof (*k));
    int key_n;
    for (key_n = 0; key_n < nr_keys; key_n++)
      {
	for (i = 0; i < prelen * 7; i++)
	  {
	    unsigned char bit = !!((1 << (6 - (i % 7))) & pw_cur[i / 7]);
	    if (!bit)
		continue;
	    // spu_insert doesn't work as expected with an unsigned char
	    // as the first argument, so use unsigned int
	    qword q = (qword)spu_insert((unsigned int)(1 << (key_n % 32)),
		    spu_splats(0), key_n / 32);
	    k[55 - i] = spu_or(k[55 - i], q);
	  }
	pw_inc(pw_cur, prelen);
      }
    // init the suffixes with SPACEs
    for (i = prelen * 7; i < len * 7; i++)
	// SPACE is 0x20: easy to build, only bit 5 needs to be set to 1
	k[55 - i] = (qword)spu_splats((6 - (i % 7) == 5) ? ~0 : 0);
    return 1;
}

void unixcrypt(const qword *k, qword *p, qword *c)
{
    qword saved_regs[48];
    qword *tmp;
    unsigned i;
    // do the 25 DES encryptions
    for (i = 0; i < 25; i++)
      {
	descrypt(p, c, k, saved_regs);
	tmp = p;
	p = c;
	c = tmp;
      }
}

void key_2_str(unsigned long long key, char *dest, unsigned dlen)
{
    unsigned i;
    for (i = 0; i < 8 && i < dlen; i++)
	dest[i] = (key >> ((7 - i) * 7)) & 0x7f;
    if (dlen)
	dest[(i < dlen) ? i : dlen - 1] = 0;
}

void compare(const qword *k, qword *c, const qword *c_to_bruteforce)
{
    int i;
    vector unsigned int result = spu_splats((unsigned)~0);
    for (i = 0; i < 64; i++)
	result = spu_and(result,
		(vector unsigned int)spu_eqv(c[i], c_to_bruteforce[i]));
    // count leading zeros in each of the 32-bit words
    result = spu_cntlz(result);
    // see if one of the counts is != 32
    for (i = 0; i < 4; i++)
	if (32 != spu_extract(result, i))
	  {
	    int bit = i * 32 + (31 - spu_extract(result, i));
	    unsigned long long keyval = extract_bit(k, 56, bit);
	    char pw[9];
	    key_2_str(keyval, pw, sizeof (pw));
	    printf("Found password (length %zd): %s\n", strlen(pw), pw);
	    printf("Hex dump:");
	    for (i = 0; pw[i]; i++)
		printf(" %02x", pw[i]);
	    printf("\n");
	    exit(0);
	  }
}

/*
 * Set the char pointed to by k[*i] to SPACE (*i should be pointing
 * to the char's less significant bit.
 */
void k_reset_char(qword *k, int *i)
{
    int j;
    for (j = 0; j < 7; j++, (*i)--)
	k[55 - *i] = (qword)spu_splats((j == 5) ? ~0 : 0);
}

/*
 * Increment the pw suffixes in {@code k}. The length of the passwds
 * is {@code len}. The length of the suffixes is {@code len - prelen}.
 *
 * Return 1 iff the suffixes have been incremented successfully, 0 if
 * they would have overflown.
 */
int k_increment_pw_suffix(qword *k, unsigned len, unsigned prelen,
	unsigned *counter)
{
    if (prelen > len)
	fprintf(stderr, "*bug*: prelen %i expected to be <= len %i\n",
		prelen, len);
    // zero-length suffixes can't be incremented
    if (!(len - prelen))
	return 0;
    (*counter)++;
    unsigned c2 = *counter;
    int i = (7 * len) - 1;
    while (i >= 7 * (int)prelen)
      {
	k[55 - i] = spu_xor(k[55 - i], (qword)spu_splats(~0));
	if (si_to_int(k[55 - i])) 
	  {
	    // if one of the suffix chars is 0x7f, reset it to 0x20
	    // and increment the next one
	    if (!(c2 % ('~' - ' ' + 1)))
	      {
		k_reset_char(k, &i);
		c2 /= ('~' - ' ' + 1);
		continue; // don't ret, don't decrement i
	      }
	    return 1;
	  }
	i--;
      }
    // going to overflow
    return 0;
}

/*
 * Tries all the passwords of length {@code len} starting from
 * password {@code pw_cur}, up to and including {@code pw_end}.
 * And increments {@code pw_cur} to the next password to try.
 * The hash to bruteforce is {@code c_to_bruteforce}.
 */
void break_pw_len(qword *c_to_bruteforce, char *pw_cur, const char *pw_end,
	unsigned len)
{
    char pw_stopat[9];
    if (strlen(pw_end) <= len)
      {
	strncpy(pw_stopat, pw_end, sizeof (pw_stopat));
	pw_stopat[sizeof (pw_stopat) - 1] = 0;
      }
    else
      {
	strcpy(pw_stopat, "~~~~~~~~");
	pw_stopat[len] = 0;
      }
    // pick a prefix length so that the suffix is never longer than 2 chars
    unsigned prelen = len <= 2 ? len : (len <= 4 ? 2 : len - 2);
    qword k[56];
    qword p[64];
    qword c[64];
    while (k_load_pw_prefix_space(k, pw_cur, pw_stopat, len, prelen))
      {
	unsigned counter = 0;
	do
	  {
	    memset(p, 0, sizeof (p));
	    unixcrypt(k, p, c);
	    compare(k, c, c_to_bruteforce);
	  }
	while (k_increment_pw_suffix(k, len, prelen, &counter));
      }
}

void bruteforcer(qword *c_to_bruteforce, char *pw_cur, const char *pw_end)
{
    unsigned len = strlen(pw_cur);
    while (len <= strlen(pw_end))
      {
	printf("Bruteforcing password length %u\n", len);
	break_pw_len(c_to_bruteforce, pw_cur, pw_end, len);
	len++;
      }
    printf("Password not found.\n");
}

void init_c_to_bruteforce(qword *c, const char *hash)
{
    unsigned salt;
    unsigned long long ciph;
    str_2_salt_ciph(hash, &salt, &ciph);
    unsigned i;
    if (salt != HARDCODED_SALT)
      {
	char hs_str[14];
	salt_ciph_2_str(hs_str, sizeof (hs_str), HARDCODED_SALT, 0);
	hs_str[2] = 0;
	fprintf(stderr,
		"This program is compiled to crack hashes starting "
		"with %s (salt %u).\n"
		"To crack this hash (%s), recompile with:\n"
		"  $ ./configure --salt=%c%c (or --salt=%u)\n",
		hs_str, HARDCODED_SALT, hash, hash[0], hash[1], salt);
	exit(1);
      }
    for (i = 0; i < 64; i++)
	c[i] = (qword)spu_splats((ciph & (1ULL << i)) ? ~0 : 0);
}

void usage(const char *name)
{
    fprintf(stderr, "Usage: %s <hash> [<first-pw> [<last-pw>]]\n", name);
}

int main(int argc, char **argv)
{
    // pw_cur must be big enough to hold 9 SPACEs (indicates the end of
    // the password space)
    char pw_cur[10];
    char pw_end[9];
    qword c_to_bruteforce[64];

    // by default break passwords from length 0 to length 8 using
    // the full printable ASCII space: SPACE (0x20) to TILDE (0x7e)
    *pw_cur = 0;
    strcpy(pw_end, "~~~~~~~~");
    if (argc > 1)
	init_c_to_bruteforce(c_to_bruteforce, argv[1]);
    else
	usage(argv[0]), exit(1);
    if (argc > 2)
      {
	strncpy(pw_cur, argv[2], sizeof (pw_cur));
	pw_cur[sizeof (pw_cur) - 1] = 0;
	if (strlen(pw_cur) > 8)
	    fprintf(stderr, "<first-pw> is longer than 8 chars\n"), exit(1);
      }
    if (argc > 3)
      {
	strncpy(pw_end, argv[3], sizeof (pw_end));
	pw_end[sizeof (pw_end) - 1] = 0;
	if (strlen(pw_end) > 8)
	    fprintf(stderr, "<last-pw> is longer than 8 chars\n"), exit(1);
      }
    bruteforcer(c_to_bruteforce, pw_cur, pw_end);
    return 0;
}
