#!/usr/bin/perl -w
#
# Generate the bitslicing code (descrypt.S, that calls the S-box functions),
# from the initial and final permutations, E-box, P-box, and key schedule
# definitions for DES.

use strict;

# For a description of the table format, see also
# http://en.wikipedia.org/wiki/DES_supplementary_material

# bit N (0..63) of the output is taken from bit $$ip[N]-1 (0..63) of the input
my $ip = [qw/
  58  50  42  34  26  18  10  2
  60  52  44  36  28  20  12  4
  62  54  46  38  30  22  14  6
  64  56  48  40  32  24  16  8
  57  49  41  33  25  17  9  1
  59  51  43  35  27  19  11  3
  61  53  45  37  29  21  13  5
  63  55  47  39  31  23  15  7
/];

# bit N (0..63) of the output is taken from bit $$fp[N]-1 (0..63) of the input
my $fp = [qw/
  40  8  48  16  56  24  64  32
  39  7  47  15  55  23  63  31
  38  6  46  14  54  22  62  30
  37  5  45  13  53  21  61  29
  36  4  44  12  52  20  60  28
  35  3  43  11  51  19  59  27
  34  2  42  10  50  18  58  26
  33  1  41  9  49  17  57  25
/];

# bit N (0..47) of the output is taken from bit $$e[N]-1 (0..31) of the input
my $e = [qw/
  32  1  2  3  4  5
  4  5  6  7  8  9
  8  9  10  11  12  13
  12  13  14  15  16  17
  16  17  18  19  20  21
  20  21  22  23  24  25
  24  25  26  27  28  29
  28  29  30  31  32  1
/];

# bit N (0..31) of the output is taken from bit $$p[N]-1 (0..31) of the input
my $p = [qw/
  16  7  20  21
  29  12  28  17
  1  15  23  26
  5  18  31  10
  2  8  24  14
  32  27  3  9
  19  13  30  6
  22  11  4  25
/];

# bit N (0..55) of the output is taken from bit $$pc1[N]-1 (0..63) of the input
my $pc1 = [qw/
  57  49  41  33  25  17  9
  1  58  50  42  34  26  18
  10  2  59  51  43  35  27
  19  11  3  60  52  44  36
  63  55  47  39  31  23  15
  7  62  54  46  38  30  22
  14  6  61  53  45  37  29
  21  13  5  28  20  12  4
/];

# bit N (0..47) of the output is taken from bit $$pc2[N]-1 (0..55) of the input
my $pc2 = [qw/
  14  17  11  24  1  5
  3  28  15  6  21  10
  23  19  12  4  26  8
  16  7  27  20  13  2
  41  52  31  37  47  55
  30  40  51  45  33  48
  44  49  39  56  34  53
  46  42  50  36  29  32
/];

# key schedule rotation, for rounds 0..15
my $rot = [qw/1 1 2 2 2 2 2 2 1 2 2 2 2 2 2 1/];

sub header
{
    print <<EOF;
.text
.balign 8

#include "s1.h"
#include "s2.h"
#include "s3.h"
#include "s4.h"
#include "s5.h"
#include "s6.h"
#include "s7.h"
#include "s8.h"

.global descrypt
.type global, \@function
descrypt:
	# copy c to \$62, k to \$63, saved_regs to \$61
	lr \$62,\$4
	lr \$63,\$5
	lr \$61,\$6
EOF
}

sub init_l_r
{
    my $i;
    printf "\t# \$64-\$127 will be used to store the 32 \"left\" and\n";
    printf "\t# 32 \"right\" bits used in each round, but the ABI requires\n";
    printf "\t# \$80-\$127 to be saved, so save them in saved_regs\n";
    for ($i = 0; $i < 48; $i++) {
	printf "\tstqd \$%i,%i(\$61)\n", 80 + $i, 16 * $i;
    }
    for ($i = 0; $i < 64; $i++) {
	my ($l_or_r, $offset);
	if ($i < 32) { ($l_or_r, $offset) = ("l", 0); }
	else         { ($l_or_r, $offset) = ("r", 32); }
	# $64-$95  is l0-l31
	# $96-$127 is r0-r31
	printf "\tlqd \$%i,%i(\$3)\n", 64 + $i, 16 * ($$ip[63 - $i] - 1);
    }
}

#
# Calculate where bit $bit of the key K_n (n is the feistel round
# number) comes from, from the original 56-bit key. Returns the
# bit position in the 56-bit key.
#
# $bit   0..47
# $round 0..15
#
sub bit_ks_for_round
{
    my ($bit, $round) = @_;
    my $keymsb;
    my $keylsb;
    my $i;
    for ($i = 0; $i < 56; $i++) {
	# The DES specs define the original 64-bit key as:
	#   [MSB] bit1  bit2 ...  bit7 (bit8) bit9  ... bit63 (bit64) [LSB]
	# Bits 8, 16, ..., 64 are used as parity bits. This code names
	# them starting from the LSB and by skipping the parity bits:
	#   [MSB] bit55 bit54 ... bit49       bit48 ... bit0          [LSB]
	my $x = $$pc1[$i] - 1;
	$x = 55 - ($x - int($x / 8));
	# now $x represents a bit 0..55 numbered according to our convention
	if ($i < 28) {
	    push @{$keymsb}, $x;
	} else {
	    push @{$keylsb}, $x;
	}
    }
    my $total = 0;
    for ($i = 0; $i <= $round; $i++) {
	$total += $$rot[$i];
    }
    my $x;
    for ($i = 0; $i < $total; $i++) {
	$x = shift @{$keymsb};
	push @{$keymsb}, $x;
	$x = shift @{$keylsb};
	push @{$keylsb}, $x;
    }
    my @concat = (@{$keymsb}, @{$keylsb});
    return $concat[$$pc2[$bit] - 1];
}

my $itoa64 = # 0..63 => ascii-64
"./0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";

sub salt_value
{
    my ($s) = @_;
    if ($s =~ /^\d+$/) {
	return int($s);
    } elsif ($s =~ m{^[./[:alnum:]]{2}([./[:alnum:]]{11})?$}) {
	return
	index($itoa64, substr($s, 0, 1)) +
	(index($itoa64, substr($s, 1, 1)) << 6);
    } else {
	die "cannot parse salt '$s'";
    }
}

#
# Computes how the salt perturbs
#
sub salt
{
    my ($salt, $e_bit) = @_;
    return ($salt & (1 << ($e_bit % 24))) ? ($e_bit + 24) % 48 : $e_bit;
}

sub feistel_round
{
    my ($salt, $round) = @_;
    my ($input_name, $output_name) = !($round % 2) ? qw/r l/ : qw/l r/;
    my $i;
    my $sbox;
    my $offset;
    my $offset_in;
    my $offset_out;
    # computes $p_rev which is a reverse map of $p: input bit N (0..31)
    # will be output as bit $$p_rev[N] (0..63)
    my $p_rev;
    for ($i = 0; $i < scalar(@{$p}); $i++) {
	$$p_rev[$$p[$i] - 1] = $i;
    }
    # computes $fp_rev which is a reverse map of $fp: input bit N (0..63)
    # will be output as bit $$fp_rev[N] (0..63)
    my $fp_rev;
    for ($i = 0; $i < scalar(@{$fp}); $i++) {
	$$fp_rev[$$fp[$i] - 1] = $i;
    }
    for ($sbox = 0; $sbox < 8; $sbox++) {
	my $inbit;
	$offset_in = ($input_name eq 'l') ? 0 : 32;
	$offset_out = ($output_name eq 'l') ? 0 : 32;
	printf "\t# s%i\n", $sbox + 1;
	for ($inbit = 0; $inbit < 6; $inbit++) {
	    # k is $63, load input bits in $3-$8
	    printf "\tlqd \$%i,%i(\$63)\n", $inbit + 3,
	    16 * bit_ks_for_round(($sbox + 1) * 6 - 1 - $inbit, $round);
	}
	for ($inbit = 0; $inbit < 6; $inbit++) {
	    printf "\txor \$%i,\$%i,\$%i\n", $inbit + 3, $inbit + 3,
	    64 + $offset_in +
	    $$e[salt($salt, ($sbox + 1) * 6 - 1 - $inbit)] - 1;
	}
	my $outbit;
	my @outputbits;
	# place output bits in $9-$12
	for ($outbit = 0; $outbit < 4; $outbit++) {
	    push @outputbits, sprintf "\$%i",
	    64 + $offset_out + $$p_rev[($sbox + 1) * 4 - 1 - $outbit];
	}
	# execute s[1-8]
	printf "\tS%i(%s)\n", $sbox + 1, join(", ", @outputbits);
	# in the last 2 rounds, save the output bits as the ciphertext
	if ($round >= 14) {
	    $offset = ($output_name eq 'r') ? 0 : 32;
	    #for ($outbit = 0; $outbit < 4; $outbit++) {
	    for ($outbit = 3; $outbit >= 0; $outbit--) {
		my $bit_nr_after_p = $$p_rev[($sbox + 1) * 4 - 1 - $outbit];
		# store output bits in c[] ($62)
		printf "\tstqd \$%i,%i(\$62)\n",
		64 + (($output_name eq 'r') ? 32 : 0) + $bit_nr_after_p,
		16 * ($$fp_rev[63 - $offset - $bit_nr_after_p]);
	    }
	}
    }
}

sub feistel_rounds
{
    my ($salt) = @_;
    my $round;
    for ($round = 0; $round < 16; $round++) {
	printf "\n";
	feistel_round($salt, $round);
    }
}

sub footer
{
    # restore the 48 regs $80-$127 from saved_regs
    my $i;
    printf "\n\t# restore the regs\n";
    for ($i = 0; $i < 48; $i++) {
	printf "\tlqd \$%i,%i(\$61)\n", 80 + $i, 16 * $i;
    }
    print "\tbi \$lr\n";
}

sub usage
{
    print STDERR "Usage: $0 <hash-or-salt>\n";
}

sub main
{
    if (scalar(@ARGV) != 1) {
	usage();
	exit(1);
    }
    if (!defined($ARGV[0])) {
	usage();
	exit(1);
    }
    my $salt = salt_value($ARGV[0]);
    header();
    init_l_r();
    feistel_rounds($salt);
    footer();
}

main();
