#include <getopt.h>
#include <limits.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "sboxes.h"

int opt_verbose = 0;
int opt_c = 0;

// operator
enum e_op {
    OP_INPUT, // one of the input bits
    OP_CONST,
    OP_NOT,
    OP_AND, OP_NAND, OP_ANDC,
    OP_OR,  OP_NOR,  OP_ORC,
    OP_XOR, OP_EQV,
    OP_SELB,
};
typedef enum e_op	t_op;

// gate
struct s_gate {
    t_op		op;
    int			ins[3]; // up to 3 inputs for tertiary ops, can be -1
    unsigned long long	tt;
};
typedef struct s_gate	t_gate;

/*
 * circuit
 *
 * bit_in        maps input bit numbers to the first 6 gates
 * bit_out       maps output bit numbers to gate indices
 */
struct s_circuit {
    unsigned		bit_in[6];
    int			bit_out[4];
    t_gate		*gates;
    unsigned		alloc;
    unsigned		len;
};
typedef struct s_circuit	t_circuit;

/*
 * Logs a text string.
 */
void lg(char *fmt, ...)
{
    va_list	ap;

    if (!opt_verbose)
	return ;
    va_start(ap, fmt);
    vprintf(fmt, ap);
    va_end(ap);
}

/*
 * Do the truth tables match according to the don't-cares ?
 */
int ttmatch(unsigned long long a, unsigned long long b,
	unsigned long long dontcare)
{
    return (a & dontcare) == (b & dontcare);
}

/*
 * Implements the Cell BE SPU selb instruction. Schematically: !rc ? ra : rb.
 */
unsigned long long opselb(unsigned long long ra, unsigned long long rb,
	unsigned long long rc)
{
    return ((~rc) & ra) | (rc & rb);
}

void gen_source_c(int sboxidx, t_circuit *c)
{
    printf("void s%i(int i0, int i1, int i2, int i3, int i4, int i5,\n"
	    "  int *o0, int *o1, int *o2, int *o3) {\n", sboxidx + 1);
    unsigned i = 0;
    while (i < 6) {
	printf("  int v%i = i%i;\n", c->bit_in[i], i);
	i++;
    }
    while (i < c->len)
      {
	t_gate *g = c->gates + i;
	switch (g->op) {
	    case OP_INPUT:
		fprintf(stderr, "gen_source_c: got OP_INPUT !?\n");
		exit(1);
	    case OP_CONST:
		printf("  int v%i = %i;\n",
			i, (int)g->tt); break;
	    case OP_NOT:
		printf("  int v%i = ~v%i;\n",
			i, g->ins[0]); break;
	    case OP_AND:
		printf("  int v%i = v%i & v%i;\n",
			i, g->ins[0], g->ins[1]); break;
	    case OP_NAND:
		printf("  int v%i = ~(v%i & v%i);\n",
			i, g->ins[0], g->ins[1]); break;
	    case OP_ANDC:
		printf("  int v%i = v%i & ~v%i;\n",
			i, g->ins[0], g->ins[1]); break;
	    case OP_OR:
		printf("  int v%i = v%i | v%i;\n",
			i, g->ins[0], g->ins[1]); break;
	    case OP_NOR:
		printf("  int v%i = ~(v%i | v%i);\n",
			i, g->ins[0], g->ins[1]); break;
	    case OP_ORC:
		printf("  int v%i = v%i | ~v%i;\n",
			i, g->ins[0], g->ins[1]); break;
	    case OP_XOR:
		printf("  int v%i = v%i ^ v%i;\n",
			i, g->ins[0], g->ins[1]); break;
	    case OP_EQV:
		printf("  int v%i = ~(v%i ^ v%i);\n",
			i, g->ins[0], g->ins[1]); break;
	    case OP_SELB:
		printf("  int v%i = ((~v%i) & v%i) | (v%i & v%i);\n",
			i, g->ins[2], g->ins[0], g->ins[2], g->ins[1]); break;
	    default:
		fprintf(stderr, "gen_source_c: unknown op: %i\n", g->op);
		exit(1);
	}
	i++;
      }
    printf("  *o0 = v%i;\n  *o1 = v%i;\n  *o2 = v%i;\n  *o3 = v%i;\n}\n",
	    c->bit_out[0], c->bit_out[1], c->bit_out[2], c->bit_out[3]);
}

/*
 * Creates an empty circuit.
 */
void c_create(t_circuit *c)
{
    unsigned initial = 8;
    c->gates = malloc(initial * sizeof (*c->gates));
    if (!c->gates) {
	perror("c_create: malloc");
	exit(1);
    }
    for (int i = 0; i < 6; i++)
	c->bit_in[i] = i;
    c->bit_out[0] = c->bit_out[1] = c->bit_out[2] = c->bit_out[3] = -1;
    c->alloc = initial;
    c->len = 0;
}

/*
 * Destroys a circuit.
 */
void c_destroy(t_circuit *c)
{
    if (c->gates)
	free(c->gates);
    c->gates = NULL;
}

/*
 * Converts a gates[] index to an input bit number, or -1.
 */
int c_toinputbit(t_circuit *c, unsigned idx)
{
    for (int bit = 0; bit < 6; bit++)
	if (c->bit_in[bit] == idx)
	    return bit;
    return -1;
}

/*
 * Copies a circuit.
 */
void c_copy(t_circuit *dst, t_circuit *src)
{
    if (dst->gates)
	free(dst->gates);
    dst->gates = malloc(src->len * sizeof (*dst->gates));
    if (!dst->gates) {
	perror("c_copy: malloc");
	exit(1);
    }
    memcpy(dst->gates, src->gates, src->len * sizeof (*dst->gates));
    memcpy(dst->bit_in, src->bit_in, sizeof (src->bit_in));
    memcpy(dst->bit_out, src->bit_out, sizeof (src->bit_out));
    dst->alloc = src->len;
    dst->len = src->len;
}

/*
 * Adds a gate to a circuit and returns its index.
 */
int c_add(t_circuit *c, t_op op, int in0, int in1, int in2,
	unsigned long long tt)
{
    int	i;
    if (c->len + 1 > c->alloc) {
	unsigned newalloc = c->alloc * 2;
	c->gates = realloc(c->gates, newalloc * sizeof (*c->gates));
	if (!c->gates) {
	    perror("c_add: realloc:");
	    exit(1);
	}
	c->alloc = newalloc;
    }
    i = c->len;
    c->gates[i].op = op;
    c->gates[i].ins[0] = in0;
    c->gates[i].ins[1] = in1;
    c->gates[i].ins[2] = in2;
    c->gates[i].tt = tt;
    c->len++;
    return i;
}

/*
 * Returns the index of the gate returning the given tt/dontcare, or -1.
 */
int c_contains(t_circuit *c, unsigned long long tt, unsigned long long dontcare)
{
    // starts searching from the end (most recent added gates)
    // TODO: use a more efficient structure (hash table)
    for (int i = c->len - 1; i >= 0; i--)
	if (ttmatch(c->gates[i].tt, tt, dontcare))
	    return i;
    return -1;
}

/*
 * Dump the string repr of a gate.
 */
char *gatedump(char *buf, unsigned len, t_circuit *c, unsigned i)
{
    t_gate *g = c->gates + i;
    switch (g->op) {
	case OP_INPUT:
	    snprintf(buf, len, "%4i: 0x%016llx (i%i)",
		    i, g->tt, c_toinputbit(c, i)); break;
	case OP_CONST:
	    snprintf(buf, len, "%4i: 0x%016llx (const %i)",
		    i, g->tt, (int)g->tt); break;
	case OP_NOT:
	    snprintf(buf, len, "%4i: 0x%016llx not %i",
		    i, g->tt, g->ins[0]); break;
	case OP_AND:
	    snprintf(buf, len, "%4i: 0x%016llx and %i,%i",
		    i, g->tt, g->ins[0], g->ins[1]); break;
	case OP_NAND:
	    snprintf(buf, len, "%4i: 0x%016llx nand %i,%i",
		    i, g->tt, g->ins[0], g->ins[1]); break;
	case OP_ANDC:
	    snprintf(buf, len, "%4i: 0x%016llx andc %i,%i",
		    i, g->tt, g->ins[0], g->ins[1]); break;
	case OP_OR:
	    snprintf(buf, len, "%4i: 0x%016llx or %i,%i",
		    i, g->tt, g->ins[0], g->ins[1]); break;
	case OP_NOR:
	    snprintf(buf, len, "%4i: 0x%016llx nor %i,%i",
		    i, g->tt, g->ins[0], g->ins[1]); break;
	case OP_ORC:
	    snprintf(buf, len, "%4i: 0x%016llx orc %i,%i",
		    i, g->tt, g->ins[0], g->ins[1]); break;
	case OP_XOR:
	    snprintf(buf, len, "%4i: 0x%016llx xor %i,%i",
		    i, g->tt, g->ins[0], g->ins[1]); break;
	case OP_EQV:
	    snprintf(buf, len, "%4i: 0x%016llx eqv %i,%i",
		    i, g->tt, g->ins[0], g->ins[1]); break;
	case OP_SELB:
	    snprintf(buf, len, "%4i: 0x%016llx selb %i,%i,%i",
		    i, g->tt, g->ins[0], g->ins[1], g->ins[2]); break;
	default:
	    fprintf(stderr, "gatedump: unknown op: %i\n", g->op);
	    exit(1);
    }
    return buf;
}

/*
 * Prints a textual repr of the circuit.
 */
void c_dump(t_circuit *c)
{
    char	gate[80];
    unsigned	i;

    printf("circuit %p (%u/%u) (i0=%i,i1=%i,i2=%i,i3=%i,i4=%i,i5=%i) "
	    "(outs:%i,%i,%i,%i)\n",
	    (void *)c->gates, c->len, c->alloc,
	    c->bit_in[0], c->bit_in[1], c->bit_in[2],
	    c->bit_in[3], c->bit_in[4], c->bit_in[5],
	    c->bit_out[0], c->bit_out[1], c->bit_out[2], c->bit_out[3]);
    i = 0;
    while (i < c->len)
      {
	printf("%s\n", gatedump(gate, sizeof (gate), c, i));
	i++;
      }
}

/*
 * Returns the truth table of an S-box input bit.
 *
 * b   input bit number (from 0 being the LSB to 5 the MSB)
 *
 * Returns the truth table.
 */
unsigned long long ttinputbit(int b)
{
    switch (b) {
	case 0: return 0xaaaaaaaaaaaaaaaaULL;
	case 1: return 0xccccccccccccccccULL;
	case 2: return 0xf0f0f0f0f0f0f0f0ULL;
	case 3: return 0xff00ff00ff00ff00ULL;
	case 4: return 0xffff0000ffff0000ULL;
	case 5: return 0xffffffff00000000ULL;
	default:
		fprintf(stderr, "ttinputbit: invalid bit number: %i\n", b);
		exit(1);
    }
}

/*
 * Returns the truth table of an S-box output bit.
 *
 * sboxidx  S-box "index" (0..7)
 * b        Output bit number (from 0 being the LSB to 3 the MSB)
 */
unsigned long long ttoutputbit(int sboxidx, int b)
{
    unsigned long long	res;
    int			i;

    if (sboxidx < 0 || sboxidx > 7) {
	fprintf(stderr, "ttoutputbit: invalid S-box index: %i\n", sboxidx);
	exit(1);
    }
    if (b < 0 || b > 3) {
	fprintf(stderr, "ttoutputbit: invalid bit number: %i\n", b);
	exit(1);
    }
    res = 0;
    for (i = 0; i < 64; i++)
	if (sboxes[sboxidx][i] & (1 << b))
	    res |= (1ULL << i);
    return res;
}

/*
 * From the circuit obtained so far, adds gates to it to obtain the truth
 * table specified by wanted/dontcare.
 *
 * Returns the index of the gate in the circuit producing the given
 * truth table. Or -1 if it is not possible to generate a circuit
 * with a length less than shortestlen.
 */
int get_crackin(t_circuit *c, unsigned long long wanted,
	unsigned long long dontcare, int nextinput, unsigned shortestlen)
{
    unsigned long long	res, r2;
    int			idx0;
    int			idx1;
    unsigned		i, j, k;

    //lg("iteration\n");
    (void)k;
    (void)r2;
    // got it already
    if (-1 != (idx0 = c_contains(c, wanted, dontcare)))
	return idx0;
    // if we don't have it and we are already as long as the shortest
    // circuit found so far, return
    if (shortestlen && c->len >= shortestlen)
	return -1;
    // constant
    if (ttmatch(0, wanted, dontcare)) {
	fprintf(stderr, "returning 0 const\n");
	return c_add(c, OP_CONST, -1, -1, -1, 0);
    }
    if (ttmatch(ULLONG_MAX, wanted, dontcare)) {
	fprintf(stderr, "returning 1 const\n");
	return c_add(c, OP_CONST, -1, -1, -1, ULLONG_MAX);
    }
    // unary gates
    for (i = 0; i < c->len; i++)
      {
	res = ~c->gates[i].tt;
	if (ttmatch(res, wanted, dontcare))
	    return c_add(c, OP_NOT, i, -1, -1, res);
      }
    // binary gates
    for (i = 0; i < c->len; i++)
	for (j = 0; j < c->len; j++)
	  {
	    res = c->gates[i].tt & c->gates[j].tt;
	    if (ttmatch(res, wanted, dontcare))
		return c_add(c, OP_AND, i, j, -1, res);
	    res = ~(c->gates[i].tt & c->gates[j].tt);
	    if (ttmatch(res, wanted, dontcare))
		return c_add(c, OP_NAND, i, j, -1, res);
	    res = c->gates[i].tt & ~c->gates[j].tt;
	    if (ttmatch(res, wanted, dontcare))
		return c_add(c, OP_ANDC, i, j, -1, res);
	    res = c->gates[i].tt | c->gates[j].tt;
	    if (ttmatch(res, wanted, dontcare))
		return c_add(c, OP_OR, i, j, -1, res);
	    res = ~(c->gates[i].tt | c->gates[j].tt);
	    if (ttmatch(res, wanted, dontcare))
		return c_add(c, OP_NOR, i, j, -1, res);
	    res = c->gates[i].tt | ~c->gates[j].tt;
	    if (ttmatch(res, wanted, dontcare))
		return c_add(c, OP_ORC, i, j, -1, res);
	    res = c->gates[i].tt ^ c->gates[j].tt;
	    if (ttmatch(res, wanted, dontcare))
		return c_add(c, OP_XOR, i, j, -1, res);
	    res = ~(c->gates[i].tt ^ c->gates[j].tt);
	    if (ttmatch(res, wanted, dontcare))
		return c_add(c, OP_EQV, i, j, -1, res);
	  }
    // ternary functions made of various gates
    for (i = 0; i < c->len; i++)
	for (j = 0; j < c->len; j++)
	  for (k = 0; k < c->len; k++)
	    {
#define TERNARY_EVAL(N1, OP1_C, OP1_N, N2, OP2_C, OP2_N) \
   res = N2((r2 = N1(c->gates[i].tt OP1_C c->gates[j].tt)) OP2_C c->gates[k].tt);\
   if (ttmatch(res, wanted, dontcare)) \
      return c_add(c,OP2_N,c_add(c,OP1_N,i,j,-1,r2),k,-1,res);
	      TERNARY_EVAL(, &, OP_AND, , &, OP_AND)
	      TERNARY_EVAL(, &, OP_AND, ~, &, OP_NAND)
	      TERNARY_EVAL(, &, OP_AND, , &~, OP_ANDC)
	      TERNARY_EVAL(, &, OP_AND, , |, OP_OR)
	      TERNARY_EVAL(, &, OP_AND, ~, |, OP_NOR)
	      TERNARY_EVAL(, &, OP_AND, , |~, OP_ORC)
	      TERNARY_EVAL(, &, OP_AND, , ^, OP_XOR)
	      TERNARY_EVAL(, &, OP_AND, , ^~, OP_EQV)
	      TERNARY_EVAL(~, &, OP_NAND, , &, OP_AND)
	      TERNARY_EVAL(~, &, OP_NAND, ~, &, OP_NAND)
	      TERNARY_EVAL(~, &, OP_NAND, , &~, OP_ANDC)
	      TERNARY_EVAL(~, &, OP_NAND, , |, OP_OR)
	      TERNARY_EVAL(~, &, OP_NAND, ~, |, OP_NOR)
	      TERNARY_EVAL(~, &, OP_NAND, , |~, OP_ORC)
	      TERNARY_EVAL(~, &, OP_NAND, , ^, OP_XOR)
	      TERNARY_EVAL(~, &, OP_NAND, , ^~, OP_EQV)
	      TERNARY_EVAL(, &~, OP_ANDC, , &, OP_AND)
	      TERNARY_EVAL(, &~, OP_ANDC, ~, &, OP_NAND)
	      TERNARY_EVAL(, &~, OP_ANDC, , &~, OP_ANDC)
	      TERNARY_EVAL(, &~, OP_ANDC, , |, OP_OR)
	      TERNARY_EVAL(, &~, OP_ANDC, ~, |, OP_NOR)
	      TERNARY_EVAL(, &~, OP_ANDC, , |~, OP_ORC)
	      TERNARY_EVAL(, &~, OP_ANDC, , ^, OP_XOR)
	      TERNARY_EVAL(, &~, OP_ANDC, , ^~, OP_EQV)
	      TERNARY_EVAL(, |, OP_OR, , &, OP_AND)
	      TERNARY_EVAL(, |, OP_OR, ~, &, OP_NAND)
	      TERNARY_EVAL(, |, OP_OR, , &~, OP_ANDC)
	      TERNARY_EVAL(, |, OP_OR, , |, OP_OR)
	      TERNARY_EVAL(, |, OP_OR, ~, |, OP_NOR)
	      TERNARY_EVAL(, |, OP_OR, , |~, OP_ORC)
	      TERNARY_EVAL(, |, OP_OR, , ^, OP_XOR)
	      TERNARY_EVAL(, |, OP_OR, , ^~, OP_EQV)
	      TERNARY_EVAL(~, |, OP_NOR, , &, OP_AND)
	      TERNARY_EVAL(~, |, OP_NOR, ~, &, OP_NAND)
	      TERNARY_EVAL(~, |, OP_NOR, , &~, OP_ANDC)
	      TERNARY_EVAL(~, |, OP_NOR, , |, OP_OR)
	      TERNARY_EVAL(~, |, OP_NOR, ~, |, OP_NOR)
	      TERNARY_EVAL(~, |, OP_NOR, , |~, OP_ORC)
	      TERNARY_EVAL(~, |, OP_NOR, , ^, OP_XOR)
	      TERNARY_EVAL(~, |, OP_NOR, , ^~, OP_EQV)
	      TERNARY_EVAL(, |~, OP_ORC, , &, OP_AND)
	      TERNARY_EVAL(, |~, OP_ORC, ~, &, OP_NAND)
	      TERNARY_EVAL(, |~, OP_ORC, , &~, OP_ANDC)
	      TERNARY_EVAL(, |~, OP_ORC, , |, OP_OR)
	      TERNARY_EVAL(, |~, OP_ORC, ~, |, OP_NOR)
	      TERNARY_EVAL(, |~, OP_ORC, , |~, OP_ORC)
	      TERNARY_EVAL(, |~, OP_ORC, , ^, OP_XOR)
	      TERNARY_EVAL(, |~, OP_ORC, , ^~, OP_EQV)
	      TERNARY_EVAL(, ^, OP_XOR, , &, OP_AND)
	      TERNARY_EVAL(, ^, OP_XOR, ~, &, OP_NAND)
	      TERNARY_EVAL(, ^, OP_XOR, , &~, OP_ANDC)
	      TERNARY_EVAL(, ^, OP_XOR, , |, OP_OR)
	      TERNARY_EVAL(, ^, OP_XOR, ~, |, OP_NOR)
	      TERNARY_EVAL(, ^, OP_XOR, , |~, OP_ORC)
	      TERNARY_EVAL(, ^, OP_XOR, , ^, OP_XOR)
	      TERNARY_EVAL(, ^, OP_XOR, , ^~, OP_EQV)
	      TERNARY_EVAL(, ^~, OP_EQV, , &, OP_AND)
	      TERNARY_EVAL(, ^~, OP_EQV, ~, &, OP_NAND)
	      TERNARY_EVAL(, ^~, OP_EQV, , &~, OP_ANDC)
	      TERNARY_EVAL(, ^~, OP_EQV, , |, OP_OR)
	      TERNARY_EVAL(, ^~, OP_EQV, ~, |, OP_NOR)
	      TERNARY_EVAL(, ^~, OP_EQV, , |~, OP_ORC)
	      TERNARY_EVAL(, ^~, OP_EQV, , ^, OP_XOR)
	      TERNARY_EVAL(, ^~, OP_EQV, , ^~, OP_EQV)
	    }
    // MUX gates
    if (nextinput < 0 || nextinput > 5) {
	fprintf(stderr, "nextinput %i doesn't make sense\n", nextinput);
	exit(1);
    }
    unsigned long long maskif1 = ttinputbit(c_toinputbit(c, nextinput));
    idx1 = get_crackin(c, wanted, dontcare & maskif1, nextinput + 1, shortestlen);
    if (idx1 == -1)
	return -1;
    idx0 = get_crackin(c, wanted, dontcare & ~maskif1, nextinput + 1, shortestlen);
    if (idx0 == -1)
	return -1;
    // don't bother adding the gate if we have already equalled the shortest circuit
    if (shortestlen && c->len >= shortestlen)
	return -1;
    return c_add(c, OP_SELB, idx0, idx1, nextinput,
	    opselb(c->gates[idx0].tt, c->gates[idx1].tt,
		c->gates[nextinput].tt));
}

/*
 * Determines the best gate circuit for one S-box.
 */
int gencircuit(int sboxidx, t_circuit *c, unsigned shortestlen, int outmap[4])
{
    unsigned long long	ttout[4];
    int			outbit;
    int			i;
    int			err;

    // initializes the 4 final truth tables we want to obtain (output bits)
    for (i = 0; i < 4; i++)
	ttout[i] = ttoutputbit(sboxidx, i);
    i = 0;
    while (i < 4) {
	outbit = outmap[i];
	if (0 && opt_verbose)
	    fprintf(stderr, "Looking for bit %i's tt %016llx\n",
		    outbit, ttout[outbit]);
	err = get_crackin(c, ttout[outbit], ULLONG_MAX, 0, shortestlen);
	if (err == -1)
	    return -1;
	c->bit_out[outbit] = err;
	i++;
    }
    return 0;
}

void choosecircuit(int sboxidx)
{
    t_circuit		c;
    t_circuit		cbest;
    unsigned		shortestlen = 0;
    int			counter = 0;
    int			err;
    int			outmap[][4] = {
	  {0,1,2,3}, {0,1,3,2}, {0,2,1,3}, {0,2,3,1}, {0,3,1,2}, {0,3,2,1},
	  {1,0,2,3}, {1,0,3,2}, {1,2,0,3}, {1,2,3,0}, {1,3,0,2}, {1,3,2,0},
	  {2,0,1,3}, {2,0,3,1}, {2,1,0,3}, {2,1,3,0}, {2,3,0,1}, {2,3,1,0},
	  {3,0,1,2}, {3,0,2,1}, {3,1,0,2}, {3,1,2,0}, {3,2,0,1}, {3,2,1,0},
    };
    const int		nrpreloaded = 6;

    printf("Bruteforcing circuit for S-box %i\n", sboxidx + 1);
    c_create(&cbest);
    for (int first = 0; first < 6; first++)
	for (int second = 0; second < 6; second++)
	    if (second != first)
		for (int third = 0; third < 6; third++)
		    if (third != first && third != second)
			for (int fourth = 0; fourth < 6; fourth++)
			    if (fourth != third && fourth != second &&
				    fourth != first)
				for (int fifth = 0; fifth < 6; fifth++)
				    if (fifth != fourth && fifth != third &&
					    fifth != second && fifth != first)
					for (int sixth = 0; sixth < 6; sixth++)
					    if (sixth != fifth &&
						    sixth != fourth &&
						    sixth != third &&
						    sixth != second &&
						    sixth != first)
						for (unsigned outorder = 0;
							outorder < sizeof (outmap) /
							sizeof (*outmap);
							outorder++)
						  {
						    if (!(counter % 7))
							fprintf(stderr, "\rattempt %i/%i (best gate count: %u)",
							    counter,
							    // 6! (for the 6 for-loops) *
							    // 4! (nr of elms in outmap)
							    // = 17280
							    17280,
							    !shortestlen ? 0 :
							    shortestlen - nrpreloaded);
						    c_create(&c);
						    c.bit_in[0] = first;
						    c.bit_in[1] = second;
						    c.bit_in[2] = third;
						    c.bit_in[3] = fourth;
						    c.bit_in[4] = fifth;
						    c.bit_in[5] = sixth;
						    // preload the 6 initial truth tables we have (for the 6 input bits)
						    // nrpreloaded must be adjusted to reflect the nr of preloaded values
						    c_add(&c, OP_INPUT, -1, -1, -1, ttinputbit(c_toinputbit(&c, 0)));
						    c_add(&c, OP_INPUT, -1, -1, -1, ttinputbit(c_toinputbit(&c, 1)));
						    c_add(&c, OP_INPUT, -1, -1, -1, ttinputbit(c_toinputbit(&c, 2)));
						    c_add(&c, OP_INPUT, -1, -1, -1, ttinputbit(c_toinputbit(&c, 3)));
						    c_add(&c, OP_INPUT, -1, -1, -1, ttinputbit(c_toinputbit(&c, 4)));
						    c_add(&c, OP_INPUT, -1, -1, -1, ttinputbit(c_toinputbit(&c, 5)));
						    err = gencircuit(sboxidx, &c, shortestlen, outmap[outorder]);
						    if (!shortestlen || (err != -1 && c.len < shortestlen)) {
							if (!shortestlen && err == -1) {
							    fprintf(stderr, "err is -1 and shortestlen is 0 !?\n");
							    exit(1);
							}
							c_copy(&cbest, &c);
							shortestlen = c.len;
						    }
						    c_destroy(&c);
						    counter++;
						  }
    fprintf(stderr, "\n");
    printf("Achieved %i gates for S-box %i\n",
	    cbest.len - nrpreloaded, sboxidx + 1);
    if (opt_verbose)
	c_dump(&cbest);
    if (opt_c)
	gen_source_c(sboxidx, &cbest);
}

void usage(const char *name)
{
    fprintf(stderr, "Usage: %s [OPTS] <N>...\n"
	    "Where <N>... are the S-boxes to compute the circuit for (1..8)\n\n"
	    "Options:\n"
	    "  -v  Output verbose result\n"
	    "  -c  Output C source code\n", name);
}

int main(int argc, char **argv)
{
    int		sboxidx;
    signed char	c;

    while (-1 != (c = getopt(argc, argv, "cv")))
      {
	switch (c) {
	    case 'v':
		opt_verbose = 1;
		break;
	    case 'c':
		opt_c = 1;
		break;
	    default:
		usage(argv[0]);
		exit(1);
		break;
	}
      }
    if (optind == argc) {
	usage(argv[0]);
	exit(1);
    }
    while (argv[optind]) {
	sboxidx = strtol(argv[optind], NULL, 0) - 1;
	if (sboxidx < 0 || sboxidx > 7) {
	    fprintf(stderr, "Invalid S-box number: %i\n", sboxidx + 1);
	    exit(1);
	}
	choosecircuit(sboxidx);
	optind++;
    }
    return 0;
}
