#!/usr/bin/perl -w
#
# From the S-box definitions, generates C language tables.

use strict;

# S-boxes definitions. For a description of the table format, see
# http://www.aci.net/kalliste/des.htm
my $s1 = "
     14  4  13  1   2 15  11  8   3 10   6 12   5  9   0  7
      0 15   7  4  14  2  13  1  10  6  12 11   9  5   3  8
      4  1  14  8  13  6   2 11  15 12   9  7   3 10   5  0
     15 12   8  2   4  9   1  7   5 11   3 14  10  0   6 13
";
my $s2 = "
     15  1   8 14   6 11   3  4   9  7   2 13  12  0   5 10
      3 13   4  7  15  2   8 14  12  0   1 10   6  9  11  5
      0 14   7 11  10  4  13  1   5  8  12  6   9  3   2 15
     13  8  10  1   3 15   4  2  11  6   7 12   0  5  14  9
";
my $s3 = "
     10  0   9 14   6  3  15  5   1 13  12  7  11  4   2  8
     13  7   0  9   3  4   6 10   2  8   5 14  12 11  15  1
     13  6   4  9   8 15   3  0  11  1   2 12   5 10  14  7
      1 10  13  0   6  9   8  7   4 15  14  3  11  5   2 12
";
my $s4 = "
      7 13  14  3   0  6   9 10   1  2   8  5  11 12   4 15
     13  8  11  5   6 15   0  3   4  7   2 12   1 10  14  9
     10  6   9  0  12 11   7 13  15  1   3 14   5  2   8  4
      3 15   0  6  10  1  13  8   9  4   5 11  12  7   2 14
";
my $s5 = "
      2 12   4  1   7 10  11  6   8  5   3 15  13  0  14  9
     14 11   2 12   4  7  13  1   5  0  15 10   3  9   8  6
      4  2   1 11  10 13   7  8  15  9  12  5   6  3   0 14
     11  8  12  7   1 14   2 13   6 15   0  9  10  4   5  3
";
my $s6 = "
     12  1  10 15   9  2   6  8   0 13   3  4  14  7   5 11
     10 15   4  2   7 12   9  5   6  1  13 14   0 11   3  8
      9 14  15  5   2  8  12  3   7  0   4 10   1 13  11  6
      4  3   2 12   9  5  15 10  11 14   1  7   6  0   8 13
";
my $s7 = "
      4 11   2 14  15  0   8 13   3 12   9  7   5 10   6  1
     13  0  11  7   4  9   1 10  14  3   5 12   2 15   8  6
      1  4  11 13  12  3   7 14  10 15   6  8   0  5   9  2
      6 11  13  8   1  4  10  7   9  5   0 15  14  2   3 12
";
my $s8 = "
     13  2   8  4   6 15  11  1  10  9   3 14   5  0  12  7
      1 15  13  8  10  3   7  4  12  5   6 11   0 14   9  2
      7 11   4  1   9 12  14  2   0  6  10 13  15  3   5  8
      2  1  14  7   4 10   8 13  15 12   9  0   3  5   6 11
";
my (@t1, @t2, @t3, @t4, @t5, @t6, @t7, @t8);

# Converts the position of a number in $s1, $s2, etc into the 6-bit input
# value it corresponds to.
sub _6bit_input_from_pos {
    my ($pos) = @_;
    return ($pos & 0x20) # first bit of input value
    | (($pos % 16) << 1) # middle 4 bits of input value
    | (($pos / 16) & 0x1); # last bit of input value
}

# Converts textual representatin $si to a table @{$ti}
#
# $si one of the variables $s1, $s2, etc textually representing an S-box
# $ti reference to tables @t1, @t2, etc conveniently representing the 6x4
#   input-to-ouput mapping
sub s2t {
    my ($si, $ti) = @_;
    my $pos = 0;
    foreach my $v (split /\s+/, $si) {
	# $si uses the original lousy format from the DES spec, so
	# we have to deal with it here
	next if $v eq '';
	$$ti[_6bit_input_from_pos($pos)] = $v;
	$pos++;
    }
}

# Converts textual representations $s1, $s2, etc to tables @t1, @t2, etc
sub sx2tx {
    eval "s2t(\$s$_, \\\@t$_)" foreach (1 .. 8);
}

# Prints tables @t1, @t2, etc
sub dump_tx {
    for my $i (1 .. 8) {
	print "s$i\n";
	for my $value (0 .. 2**6-1) {
	    printf "\t%02i->%04b", $value, eval("\$t${i}[$value]");
	    print "\n" if !(($value + 1) % 8);
	}
	print "\n";
    }
}

# Creates *.tbl files to be used by minilog
sub create_tbl_tx {
    my $fh;
    my $fname;
    for my $i (1 .. 8) {
	$fname = "s$i.tbl";
	open($fh, '>', $fname) or die "open $fname: $!";
	print $fh "table s$i\n  input a b c d e f\n  output x y z t\n";
	for my $value (0 .. 2**6-1) {
	    printf $fh "  %06b %04b\n", $value, eval("\$t${i}[$value]");
	}
	print $fh "end\n";
	close($fh) or die "close $fname: $!";
    }
}

# Prints S-boxes in C format
sub sboxes_c_format {
    print "unsigned char sboxes[8][64] = {";
    for my $i (1 .. 8) {
	print "\n  {";
	for my $value (0 .. 2**6-1) {
	    print "\n  " if !($value % 16);
	    printf " %i,", eval("\$t${i}[$value]");
	}
	print "\n  },";
    }
    print "\n};\n";
}

sub main {
    sx2tx();
    #dump_tx();
    sboxes_c_format();
    #create_tbl_tx();
}

main();
